import json, re, sys, time
import pytz
import requests

import numpy as np
import pandas as pd


from sentenai.exceptions import *
from sentenai.exceptions import handle
from sentenai.historiQL import Select, Returning
from sentenai.utils import *
from sentenai.api.stream import Stream, Event
from pandas.io.json import json_normalize


class Search(object):
    def __init__(self, client, select):
        self.client = client
        self.query = select

    def all(self):
        return ResultSet(self).all()

    def __iter__(self):
        return ResultSet(self)

    def __getitem__(self, i):
        return ResultSet(self)[i]

    def _spans(self, cursor):
        url = '{0}/query/{1}/spans'.format(self.client.host, cursor)
        retries = 0
        while retries < 3:
            try:
                return handle(self.client.session.get(url, params={'optimize': 'true'})).json()
            except:
                retries += 1
        else:
            raise SentenaiException("Server failed.")

    def _repr_html_(self):
        return "<pre>%s</pre>" % str(self.query)

    @property
    def ast(self):
        return self.query()

    def df(self):
        return ResultSet(self).df



class ResultPage(object):
    def __init__(self, search, *results, **kwargs):
        self.search = search
        self.results = results
        self.start = kwargs.get("start")
        self.stop = kwargs.get("stop")

    def __len__(self):
        return len(self.results) or 0

    def __getitem__(self, i):
        return self.results[i]

    def _repr_html_(self):
        if not self.results: return ""
        maxd = max([
            float(r.duration.total_seconds())
            for r in self.results
            if r.duration < timedelta.max])

        x = lambda r: int(round(r.duration.total_seconds() / maxd * 10)) * u"\u2588"
        df = pd.DataFrame([
            { 'Start': r.start,
              'End': r.end,
              'Duration': u"{}".format(r.duration if r.duration else 0),
              'Viz': u"{}".format(x(r)),
            } for r in self.results],
            index=range(self.start or 0, self.stop or (self.start or 0) + len(self.results))
            )
        if df.empty:
            return "_"
        else:
            return df[['Start', 'End', 'Duration', 'Viz']]._repr_html_()


class ResultSet(object):
    def __init__(self, search):
        self.search = search
        r = handle(self.search.client.session.post(
                '{0}/query'.format(search.client.host),
                json = self.search.query(),
            ))
        self.cursors = [r.headers['location']]
        self.spans = {}
        self.pos = (False, 0)

    def __getitem__(self, i):
        one = type(i) is not slice
        if one: i = slice(i,i+1)
        n = 0
        data = []
        while i.stop is None or n <= i.stop:
            try:
                c = self.cursors[n]
            except IndexError:
                try:
                    next(self)
                except StopIteration:
                    break
            else:
                if c is None and one:
                    raise IndexError
                elif c is None:
                    break
                elif c not in self.spans:
                    try:
                        next(self)
                    except StopIteration:
                        break
                else:
                    n += len(self.spans[c])
        for c in self.cursors:
            if c in self.spans:
                data.extend(self.spans[c])
            else:
                break
        if one:
            return data[0]
        else:
            return ResultPage(self.search, *data[i], start=i.start, stop=i.stop)

    def __iter__(self):
        return self

    def __len__(self):
        return sum([len(x) for x in self.spans.values() if x])

    def __next__(self):
        pg, ps = self.pos
        if pg in self.spans and len(self.spans[pg]) > ps:
            r = self.spans[pg].results[ps]
            self.pos = (pg, ps + 1)
            return r
        elif self.cursors[-1] is None:
            raise StopIteration
        else:
            cc = self.cursors[-1]
            self.pos = (cc, 0)
            r = self.search._spans(cc)
            c = r.get('cursor')
            self.cursors.append(c)
            rp = ResultPage(self.search, *[Result(self.search, **s) for s in r['spans']])
            self.spans[cc] = rp
            return next(self)

    next = __next__


    @property
    def complete(self):
        self.cursors[-1] is None


    @property
    def df(self):
        dfs = []
        for x in self[:]:
            dfs.append(x.df)
        return pd.concat(dfs, keys=range(0,len(dfs)))


    def all(self):
        while self.cursors[-1]:
            if self.cursors[-1] not in self.spans:
                r = self.search._spans(self.cursors[-1])
                c = r.get('cursor')
                rp = ResultPage(self.search, *[Result(self.search, **s) for s in r['spans']])
                self.spans[self.cursors[-1]] = rp
                self.cursors.append(c)
        return self


    def _repr_html_(self):
        data = []
        for cursor in self.cursors:
            data.extend(self.spans.get(cursor, []))
        if data:
            maxd = max([
                float(r.duration.total_seconds())
                for r in data
                if r.duration < timedelta.max])

        x = lambda r: int(round(r.duration.total_seconds() / maxd * 10)) * u"\u2588"
        df = pd.DataFrame([
            { 'Start': r.start,
              'End': r.end,
              'Duration': u"{}".format(r.duration if r.duration else 0),
              'Viz': u"{}".format(x(r)),
            } for r in data])
        if df.empty:
            return "_"
        else:
            return df[['Start', 'End', 'Duration', 'Viz']]._repr_html_()




class Result(object):
    def __init__(self, search, start, end, cursor):
        self.search = search
        self.start = cts(start) if start else None
        self.end = cts(end) if end else None
        self.events = 0
        self.cursor = cursor


    @property
    def duration(self):
        if self.end and self.start:
            return self.end - self.start
        else:
            return timedelta.max


    def _repr_html_(self):
        df = pd.DataFrame([
            { 'Start': self.start,
              'End': self.end,
              'Duration': u"{}".format(self.duration if self.duration else 0),
            }])
        if df.empty:
            return "_"
        else:
            return df[['Start', 'End', 'Duration']]._repr_html_()

    def _count(self, max_retries=1):
        if max_retries < 0:
            raise SentenaiException("Max Retries Exceeded.")
        streams = {}
        url = '{host}/query/{cursor}/events'.format(host=self.search.client.host, cursor=self.cursor)
        try:
            r = self.search.client.session.get(url, params={'limit': '0'})
            if not r.ok:
                if max_retries > 0:
                    return self._events(max_retries - 1)
                else:
                    raise SentenaiException(r.status_code)
        except:
            if max_retries > 0:
                return self._events(max_retries - 1)
            else:
                raise
        else:
            self._json = r.json()
            print(self._json)
            return 0

    def _events(self, max_retries=3):
        if max_retries < 0:
            raise SentenaiException("Max Retries Exceeded.")
        streams = {}
        url = '{host}/query/{cursor}/events'.format(host=self.search.client.host, cursor=self.cursor)
        try:
            r = self.search.client.session.get(url)
            if not r.ok:
                if max_retries > 0:
                    return self._events(max_retries - 1)
                else:
                    raise SentenaiException(r.status_code)
        except:
            if max_retries > 0:
                return self._events(max_retries - 1)
            else:
                raise
        else:
            self._json = r.json()
            return self._json

    def json(self):
        try:
            return self._json
        except:
            return self._events()

    @property
    def df(self):
        x = json_normalize([evt.json() for evt in list(self)])
        return x


    @property
    def streams(self):
        data = self.json()
        if not data:
            return []
        else:
            return [Stream(self.search.client, s['name'], {}, {}, None)
                    for s in data.get('streams', []).values()]


    def __iter__(self):
        data = self.json()
        streams = {k: Stream(self.search.client, s['name'], {}, {}, None)
                   for k, s in data.get('streams', {}).items()}
        return iter([Event(self.search.client, streams[e['stream']], e['id'], e['ts'], e['event'])
                     for e in data['events']])

    def __getitem__(self, i):
        ss = []
        if type(i) is tuple:
            for s in self.streams:
                if s.name in i:
                    ss.append(s)
        else:
            for s in self.streams:
                if s.name == i:
                    ss.append(s)
        return RView(self, ss)

class RView(object):
    def __init__(self, result, streams, head=None):
        self.result = result
        self.streams = streams
        self._head = head

    def _view(self):
        data = self.result.json()
        rdict = {k: [] for k in self.streams}
        sdict = {k: Stream(self.result.search.client, s['name'], {}, {}, None)
                   for k, s in data.get('streams', {}).items()}
        for e in data['events']:
            st = sdict.get(e['stream'])
            if st in self.streams:
                if self._head is None or len(rdict[st]) < self._head:
                    rdict[st].append(Event(self.result.search.client, st, e['id'], e['ts'], e['event']))
        return rdict

    def _repr_html_(self):
        rd = self._view()
        vs = []
        for k, v in rd.items():
            vs.append("<div><b>{}</b></div>".format(k.name) + json_normalize([x.json() for x in v])._repr_html_())

        return "<hr/>".join(vs)

    def head(self, n=20):
        return RView(self.result, self.streams, n)

    @property
    def df(self):
        return json_normalize([x.json() for x in v])



class ProjectedStream(object):
    def __init__(self, stream, projections):
        self._stream = stream
        self._projections = projections




__all__ = ["Within", "After", "Lasting", "During", "And", "Or", 'V']

from sentenai.historiQL import Within, After, Lasting, And, Or, EventPath, Par, InCircle, InPolygon

from sentenai.utils import PY3


# Python 2 Compatibility Decorator
if not PY3: import virtualtime


# Convenience Functions


def event(*args, **kwargs):
    return Switch(*args, **kwargs)



def within_distance(km, of):
    """Return all events within a given distance (in km) from a point."""
    return InCircle(of, km)


def inside_region(poly):
    """Return all events within a polygon."""
    return InPolygon(poly)

class Any(object):
    def __new__(self, *q):
        return Par("any", q)


class All(object):
    def __new__(self, *q):
        return Par("all", q)


class During(object):
    def __new__(self, *q):
        return Par("during", q)


V = EventPath()



import dateutil
import dateutil.tz
import sys
from datetime import datetime, timedelta, tzinfo

# Constants

PY3 = sys.version_info[0] == 3
LEFT, CENTER, RIGHT = range(-1, 2)
DEFAULT = None

if not PY3: import virtualtime

def py2str(cls):
    """Encode strings to utf-8 if the major version is not 3."""
    if not PY3:
        cls.__unicode__ = cls.__str__
        cls.__str__ = lambda self: self.__unicode__().encode('utf-8')
    return cls


class UTC(tzinfo):
    """A timezone class for UTC."""

    def dst(self, dt): return None

    def utcoffset(self, dt):
        """Generate a timedelta object with no offset."""
        return timedelta()


def iso8601(dt):
    """Convert a datetime object to an ISO8601 unix timestamp."""
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=UTC())
    return dt.isoformat()


def cts(ts):
    """Convert a time string to a datetime object."""
    try:
        dt = dateutil.parser.parse(ts)
        if dt.tzinfo:
            return dt
        else:
            return dt.replace(tzinfo=dateutil.tz.tzutc())
    except:
        print("invalid time: " + ts)
        return ts


def dts(obj):
    """Convert a timestring to an ISO8601 unix timestamp."""
    if isinstance(obj, datetime):
        serial = iso8601(obj)
        return serial
    else:
        return obj


def divtime(l, r):
    numerator = l.days * 3600 * 24 + l.seconds
    divisor   = r.days * 3600 * 24 + r.seconds
    return int( numerator / divisor )


DTMIN = datetime.min.replace(tzinfo=dateutil.tz.tzutc())
DTMAX = datetime.max.replace(tzinfo=dateutil.tz.tzutc())



from __future__ import print_function

import pytz
import re, sys, time, json
import requests

from datetime import timedelta
from multiprocessing.pool import ThreadPool
from threading import Lock

from sentenai.api.stream import Stream, Event
from sentenai.historiQL import EventPath, Returning, delta, Delta, Query
from sentenai.utils import *
from sentenai.exceptions import *
from sentenai.exceptions import handle

if PY3:
    string_types = str
else:
    string_types = basestring

if not PY3:
    import virtualtime
    from Queue import Queue
else:
    from queue import Queue

class Uploader(object):
    def __init__(self, client, iterator, processes=32):
        self.client = client
        self.iterator = iterator
        self.pool = ThreadPool(processes)
        self.succeeded = 0
        self.failed = 0
        self.lock = Lock()

    def start(self, progress=False):
        def process(data):
            def waits():
                yield 0
                wl = (0,1)
                while True:
                    wl = (wl[-1], sum(wl))
                    yield wl[-1]

            event = self.validate(data)
            if isinstance(event, tuple):
                return event

            wait = waits()
            while event:
                try:
                    self.client.put(**event)
                    with self.lock:
                        self.succeeded += 1
                    return None
                except AuthenticationError:
                    raise
                except Exception as e:
                    w = next(wait)
                    if w < 15: # 15 second wait limit
                        time.sleep(next(wait))
                    else:
                        with self.lock:
                            self.failed += 1
                        return (event, e)
                    """
                    if e.response.status_code == 400:
                        # probably bad JSON
                        return data
                    else:
                        time.sleep(next(wait))
                    """
        if progress:
            events = list(self.iterator)
            total  = len(events)
            def bar():
                sys.stderr.write("\r" * 60)
                sc = int(round(50 * self.succeeded / float(total)))
                fc = int(round(50 * self.failed    / float(total)))
                pd = (self.failed + self.succeeded) / float(total) * 100.
                sys.stderr.write(" [\033[92m{0}\033[91m{1}\033[0m{2}] {3:>6.2f}%   ".format( "#" * sc, "#" * fc, " " * (50 - sc - fc), pd))
                sys.stderr.flush()

            t0 = datetime.utcnow()
            data = self.pool.map_async(process, events)
            #sys.stderr.write("\n " + "-" * 62 + " \n")
            sys.stderr.write("\n {:<60} \n".format("Uploading {} objects:".format(total)))
            while not data.ready():
                bar()
                time.sleep(.1)
            else:
                bar()
            t1 = datetime.utcnow()
            sys.stderr.write("\n {:<60} ".format("Time elapsed: {}".format(t1 - t0)))
            sys.stderr.write("\n {:<60} ".format("Mean Obj/s: {:.1f}".format(float(total)/(t1 - t0).total_seconds())))
            sys.stderr.write("\n {:<60} \n\n".format("Failures: {}".format(self.failed)))
            #sys.stderr.write("\n " + "-" * 62 + " \n\n")
            sys.stderr.flush()
            #data.wait()
            data = data.get()
        else:
            data = self.pool.map(process, self.iterator)
        return { 'saved': self.succeeded, 'failed': list(filter(None, data)) }


    def validate(self, data):
        ts = data.get('ts')
        try:
            if isinstance(ts, string_types):
                ts = cts(ts)
            if not ts.tzinfo:
                ts = pytz.utc.localize(ts)
        except:
            return (data, "invalid timestamp")

        sid = data.get('stream')
        if isinstance(sid, Stream):
            strm = sid
        elif sid:
            strm = self.client.Stream(sid)
        else:
            return (data, "missing stream")

        try:
            evt = data['event']
        except KeyError:
            return (data, "missing event data")
        except Exception:
            return (data, "invalid event data")
        else:
            return {"stream": strm, "timestamp": ts, "id": data.get('id'), "event": evt}



from __future__ import print_function
import json as JSON
import pytz
from copy import copy
import re, sys, time
import requests

import numpy as np
import pandas as pd

from pandas.io.json import json_normalize
from datetime import timedelta
from functools import partial
from multiprocessing.pool import ThreadPool
from threading import Lock

from sentenai.exceptions import *
from sentenai.exceptions import handle
from sentenai.utils import *
from sentenai.historiQL import EventPath, Stream, StreamPath

BaseStream = Stream


if PY3:
    string_types = str
else:
    string_types = basestring

if not PY3:
    import virtualtime

try:
    from urllib.parse import quote
except:
    from urllib import quote



class Event(object):
    def __init__(self, client, stream, id=None, ts=None, data=None, event=None, saved=False):
        self.stream = stream
        self.id = id
        self.ts = ts if isinstance(ts, datetime) or ts is None else cts(ts)
        self.data = data or event
        self._saved = saved

    @property
    def exists(self):
        return bool(self._saved)


    def __repr__(self):
        return "Event(stream={}, id={}, ts={}, exists={})".format(self.stream.name, self.id, self.ts, self.exists)

    def _repr_html_(self):
        return '<pre>Event(\n  stream = "{}",\n  id = "{}",\n  ts = {},\n  exists = {},\n  data = {})</pre>'.format(self.stream.name, self.id, repr(self.ts), self.exists, JSON.dumps(self.data, indent=4, default=dts))


    def json(self, include_id=False, df=False):
        if df:
            d = copy(self.data)
            d['ts'] = self.ts
            return d
        elif include_id:
            return {'ts': self.ts, 'event': self.data, 'id': self.id}
        else:
            return {'ts': self.ts, 'event': self.data}

    def create(self):
        loc = self.stream.put(self.data, self.id, self.ts)
        self.id = loc
        self._saved = True
        return self

    def read(self):
        x = self.stream.read(self.id)
        self.ts = x.ts
        self.data = x.data
        self._saved = True
        return self

    def update(self):
        if not self.id:
            raise Exception("Not found")
        loc = self.stream.put(self.data, self.id, self.ts)
        self.id = loc
        self._saved = True
        return self

    def delete(self):
        self.stream.delete(self.id)
        self._saved = False


class Values(object):
    def __init__(self, stream, at, data):
        self.at = at
        self._data = data

    def __getitem__(self, i):
        if isinstance(i, StreamPath):
            nd = []
            for d in self._data:
                if d['path']._attrlist == i._attrlist:
                    return d['value']
            else:
                raise IndexError
        else:
            return self._data[i]


    def _repr_html_(self):
        df = pd.DataFrame(self._data)
        df['path'] = df['path'].apply(lambda x: ".".join(x._attrlist))
        df = df[['path', 'value', 'timestamp']]
        return df.sort_values(by='path').rename(
                index=str,
                columns={
                    'path': 'Event Path',
                    'value': 'Value'.format(self.at),
                    'timestamp': 'Updated At'
                    }
            )._repr_html_()

    def items(self):
        return [(d['path'], d['value']) for d in self._data]



class Fields(object):
    def __init__(self, fields):
        self._fields = [f for f in fields if f['path']]

    def __getitem__(self, path):
        xs = []
        for field in self._fields:
            if field._attrlist[:len(path._attrlist)] == path._attrlist:
                xs.append(field)
        return Fields(xs)

    def __repr__(self):
        return repr(self._fields)

    def __iter__(self):
        return iter(self._fields)

    def _repr_html_(self):
        df = pd.DataFrame(sorted(self._fields, key=lambda x: (x['start'], ".".join(x['path']))))
        df['path'] = df['path'].apply(lambda x: ".".join(x))
        df['start'] = df['start'].apply(cts)
        df = df[['path', 'start']]
        return df.rename(
                index=str,
                columns={
                    'path': 'Field',
                    'start': 'Added At'
                    }
            )._repr_html_()





class Stream(BaseStream):
    def __init__(self, client, name, meta, events, tz, exists, *filters):
        self._client = client
        self._events = events
        self._exists = exists
        BaseStream.__init__(self, name, meta, tz, *filters)

    def __len__(self):
        return self.stats().get('events')

    def __bool__(self):
        return self._exists
    __nonzero__ = __bool__


    def filtered(self, *filters, **kwargs):
        """Return copy of stream with additional filters.

        Keyword Argument:
            replace -- when True, copy stream while replacing filters instead of adding them.
        """
        if kwargs.get("replace", False):
            return Stream(self._client, self.name, {}, 0, self.tz, self._exists, filters)
        else:
            return Stream(self._client, self.name, {}, 0, self.tz, self._exists, tuple(self._filters) + filters)


    def __getattribute__(self, name):
        if hasattr(Stream, name) and hasattr(Stream, "_" + name):
            raise AttributeError("Cannot call this method.")
        else:
            return BaseStream.__getattribute__(self, name)

    def oldest(self):
        """Get the oldest event by timestamp in this stream.
        """
        return self._client.oldest(self)
    _oldest = oldest

    def newest(self):
        """Get the newest event by timestamp in this stream.
        """
        return self._client.newest(self)
    _newest = newest

    def fields(self):
        """Get a view of all fields in this stream."""
        return Fields(self._client.fields(self))
    _fields = fields

    def stats(self):
        """Get a dictionary of stream statistics."""
        return self._client.stats(self)
    _stats = stats

    def values(self, at=None):
        """Get current values for every field in a stream.

        Keyword Arguments:
            at -- If given a datetime for `at`, return the values at that point in
                  time instead
        """

        at = at or datetime.utcnow()
        values = self._client.values(self, at)
        values_rendered = []
        for value in values:
            # create path
            pth = self
            for segment in value['path']:
                pth = pth[segment]
            ts = cts(value['ts'])
            values_rendered.append({
                'timestamp': ts,
                'event': value['id'],
                'value': value['value'],
                'path': pth,
            })
        return Values(self, at, values_rendered)
    _values = values

    def Event(self, *args, **kwargs):
        return Event(self.client, self, *args, **kwargs)
    _Event = Event

    def healthy(self):
        """Did the last `create` or `update` of an event on
        this stream succeed. For debugging purposes."""
        resp = self.session.get("/".join([self.client.host, "streams", self.name]))
        if resp.status_code == 200:
            return resp.json().get('healthy', False)
        elif resp.status_code == 404:
            return None
        else:
            handle(resp)

    _healthy = healthy

    def destroy(self, **kwargs):
        """Delete stream.

        Keyword Argument:
            confirm -- Must be `True` to confirm destroy stream
        """
        return self._client.destroy(self, **kwargs)
    _destroy = destroy

    def delete(self, id):
        """Delete event from the stream by its unique id.

        Arguments:
           eid    -- A unique ID corresponding to an event stored within
                     the stream.
        """
        return self._client.delete(self, id)
    _delete = delete

    def get(self, id):
        """Get event as JSON.

        Arguments:
           eid    -- A unique ID corresponding to an event stored within
                     the stream.
        """
        return self._client.get(self, id)
    _get = get

    def read(self, id):
        """Get event as JSON.

        Arguments:
           eid    -- A unique ID corresponding to an event stored within
                     the stream.
        """
        k = self._client.get(self, id)
        return Event(self._client, self, data=k['event'], id=k['id'], ts=cts(k['ts']), saved=True)
    _read = read

    def fstats(self, field, start=None, end=None):
        """Get stats for a given numeric field.

           Arguments:
           field  -- A dotted field name for a numeric field in the stream.
           start  -- Optional argument indicating start time in stream for calculations.
           end    -- Optional argument indicating end time in stream for calculations.
        """
        return self._client.stats(self, field, start, end)
    _fstats = fstats

    def describe(self, field, start=None, end=None):
        """Describe a given numeric field.

           Arguments:
           field  -- A dotted field name for a numeric field in the stream.
           start  -- Optional argument indicating start time in stream for calculations.
           end    -- Optional argument indicating end time in stream for calculations.
        """
        x = self._client.stats(self, field, start, end)
        if x.get('categorical'):
            print("count\t{count}\nunique\t{unique}\ntop\t{top}\nfreq\t{freq}".format(**x['categorical']))
        else:
            p = x['numerical']
            print("count\t{}\nmean\t{:.2f}\nstd\t{:.2f}\nmin\t{}\n25%\t{}\n50%\t{}\n75%\t{}\nmax\t{}".format(
                p['count'], p['mean'], p['std'], p['min'], p['25%'], p['50%'], p['75%'], p['max']))
    _describe = describe

    def unique(self, field):
        """Get unique values for a given field.

           Arguments:
           field  -- A dotted field name for a numeric field in the stream.
           start  -- Optional argument indicating start time in stream for calculations.
           end    -- Optional argument indicating end time in stream for calculations.
        """
        return self._client.unique(self, field)
    _unique = unique

    def put(self, event, id=None, timestamp=None):
        """Put a new event into this stream.

        Arguments:
           event     -- A JSON-serializable dictionary containing an
                        event's data
           id        -- A user-specified id for the event that is unique to
                        this stream (optional)
           timestamp -- A user-specified datetime object representing the
                        time of the event. (optional)
        """
        return self._client.put(self, event, id, timestamp)
    _put = put

    def range(self, start, end):
        """Get all of a stream's events between start (inclusive) and end (exclusive).

        Arguments:
           start  -- A datetime object representing the start of the requested
                     time range.
           end    -- A datetime object representing the end of the requested
                     time range.

           Result:
           A time ordered list of all events in a stream from `start` to `end`
        """
        return StreamRange(self, start, end, self._client.range(self, start, end))
    _range = range

    def tail(self, n=5):
        """Get all of a stream's events between start (inclusive) and end (exclusive).

        Arguments:
           n      -- A max number of events to return

           Result:
           A time ordered list of all events in a stream from `start` to `end`
        """
        x = self._client.tail(self, n)
        if not x:
            return StreamRange(self, datetime.min, datetime.max, [])
        else:
            return StreamRange(self, x[0].ts, x[-1].ts, reversed(x))

    def head(self, n=5):
        """Get all of a stream's events between start (inclusive) and end (exclusive).

        Arguments:
           n      -- A max number of events to return

           Result:
           A time ordered list of all events in a stream from `start` to `end`
        """
        x = self._client.head(self, n)
        if not x:
            return StreamRange(self, datetime.min, datetime.max, [])
        else:
            return StreamRange(self, x[0].ts, x[-1].ts, x)

    def newest(self):
        """Get the most recent event in the stream."""
        return self._client.newest(self)
    _newest = newest

    def oldest(self):
        """Get the oldest event in the stream."""
        return self._client.oldest(self)
    _oldest = oldest




class StreamRange(object):
    def __init__(self, stream, start, end, events):
        self.stream = stream
        self._events = events
        self.start = start
        self.end = end


    def __iter__(self):
        return iter(self._events)

    @property
    def df(self):
        f = json_normalize([x.json(df=True) for x in self._events])
        return f.set_index('ts')

    def json(self):
        return JSON.dumps([x.json(include_id=True) for x in self._events], default=dts, indent=4)

    def _repr_html_(self):
        return self.df._repr_html_()


class StreamsView(object):
    def __init__(self, streams):
        self.streams = streams

    def _repr_html_(self):
        return pd.DataFrame([{ 'name': s.name, 'length': len(s), 'healthy': True} for s in self.streams])[['name', 'length', 'healthy']]._repr_html_()

    def __iter__(self):
        return self.streams

    def __getitem__(self, i):
        return self.streams[i]





















